@@ -2,8 +2,8 @@
 /*
  * Splashy - Read a password from splashy and output it
  * 
- * Copyright © 2008,2009 Teddy Hogeborn
- * Copyright © 2008,2009 Björn Påhlsson
+ * Copyright © 2008-2011 Teddy Hogeborn
+ * Copyright © 2008-2011 Björn Påhlsson
  * 
  * This program is free software: you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as
@@ -29,7 +29,7 @@
 				   SIG_IGN, kill(), SIGKILL */
 #include <stddef.h>		/* NULL */
 #include <stdlib.h>		/* getenv() */
-#include <stdio.h>		/* asprintf(), perror() */
+#include <stdio.h>		/* asprintf() */
 #include <stdlib.h>		/* EXIT_FAILURE, free(),
 				   EXIT_SUCCESS */
 #include <sys/types.h>		/* pid_t, DIR, struct dirent,
@@ -43,9 +43,17 @@
 				   STDOUT_FILENO, _exit(),
 				   pause() */
 #include <string.h>		/* memcmp() */
-#include <errno.h>		/* errno */
+#include <errno.h>		/* errno, EACCES, ENOTDIR, ELOOP,
+				   ENOENT, ENAMETOOLONG, EMFILE,
+				   ENFILE, ENOMEM, ENOEXEC, EINVAL,
+				   E2BIG, EFAULT, EIO, ETXTBSY,
+				   EISDIR, ELIBBAD, EPERM, EINTR,
+				   ECHILD */
+#include <error.h>		/* error() */
 #include <sys/wait.h>		/* waitpid(), WIFEXITED(),
 				   WEXITSTATUS() */
+#include <sysexits.h>		/* EX_OSERR, EX_OSFILE,
+				   EX_UNAVAILABLE */
 
 sig_atomic_t interrupted_by_signal = 0;
 int signal_received;
@@ -65,6 +73,7 @@ int main(__attribute__((unused))int argc,
   DIR *proc_dir = NULL;
   pid_t splashy_pid = 0;
   pid_t splashy_command_pid = 0;
+  int exitstatus = EXIT_FAILURE;
   
   /* Create prompt string */
   {
@@ -90,6 +99,7 @@ int main(__attribute__((unused))int argc,
     }
     if(ret == -1){
       prompt = NULL;
+      exitstatus = EX_OSERR;
       goto failure;
     }
   }
@@ -99,7 +109,23 @@ int main(__attribute__((unused))int argc,
     const char splashy_name[] = "/sbin/splashy";
     proc_dir = opendir("/proc");
     if(proc_dir == NULL){
-      perror("opendir");
+      int e = errno;
+      error(0, errno, "opendir");
+      switch(e){
+      case EACCES:
+      case ENOTDIR:
+      case ELOOP:
+      case ENOENT:
+      default:
+	exitstatus = EX_OSFILE;
+	break;
+      case ENAMETOOLONG:
+      case EMFILE:
+      case ENFILE:
+      case ENOMEM:
+	exitstatus = EX_OSERR;
+	break;
+      }
       goto failure;
     }
     for(struct dirent *proc_ent = readdir(proc_dir);
@@ -126,7 +152,8 @@ int main(__attribute__((unused))int argc,
 	char *exe_link;
 	ret = asprintf(&exe_link, "/proc/%s/exe", proc_ent->d_name);
 	if(ret == -1){
-	  perror("asprintf");
+	  error(0, errno, "asprintf");
+	  exitstatus = EX_OSERR;
 	  goto failure;
 	}
 	
@@ -138,8 +165,20 @@ int main(__attribute__((unused))int argc,
 	    free(exe_link);
 	    continue;
 	  }
-	  perror("lstat");
+	  int e = errno;
+	  error(0, errno, "lstat");
 	  free(exe_link);
+	  switch(e){
+	  case EACCES:
+	  case ENOTDIR:
+	  case ELOOP:
+	  default:
+	    exitstatus = EX_OSFILE;
+	    break;
+	  case ENAMETOOLONG:
+	    exitstatus = EX_OSERR;
+	    break;
+	  }
 	  goto failure;
 	}
 	if(not S_ISLNK(exe_stat.st_mode)
@@ -163,6 +202,7 @@ int main(__attribute__((unused))int argc,
     proc_dir = NULL;
   }
   if(splashy_pid == 0){
+    exitstatus = EX_UNAVAILABLE;
     goto failure;
   }
   
@@ -174,52 +214,61 @@ int main(__attribute__((unused))int argc,
     sigemptyset(&new_action.sa_mask);
     ret = sigaddset(&new_action.sa_mask, SIGINT);
     if(ret == -1){
-      perror("sigaddset");
+      error(0, errno, "sigaddset");
+      exitstatus = EX_OSERR;
       goto failure;
     }
     ret = sigaddset(&new_action.sa_mask, SIGHUP);
     if(ret == -1){
-      perror("sigaddset");
+      error(0, errno, "sigaddset");
+      exitstatus = EX_OSERR;
       goto failure;
     }
     ret = sigaddset(&new_action.sa_mask, SIGTERM);
     if(ret == -1){
-      perror("sigaddset");
+      error(0, errno, "sigaddset");
+      exitstatus = EX_OSERR;
       goto failure;
     }
     ret = sigaction(SIGINT, NULL, &old_action);
     if(ret == -1){
-      perror("sigaction");
+      error(0, errno, "sigaction");
+      exitstatus = EX_OSERR;
       goto failure;
     }
     if(old_action.sa_handler != SIG_IGN){
       ret = sigaction(SIGINT, &new_action, NULL);
       if(ret == -1){
-	perror("sigaction");
+	error(0, errno, "sigaction");
+	exitstatus = EX_OSERR;
 	goto failure;
       }
     }
     ret = sigaction(SIGHUP, NULL, &old_action);
     if(ret == -1){
-      perror("sigaction");
+      error(0, errno, "sigaction");
+      exitstatus = EX_OSERR;
       goto failure;
     }
     if(old_action.sa_handler != SIG_IGN){
       ret = sigaction(SIGHUP, &new_action, NULL);
       if(ret == -1){
-	perror("sigaction");
+	error(0, errno, "sigaction");
+	exitstatus = EX_OSERR;
 	goto failure;
       }
     }
     ret = sigaction(SIGTERM, NULL, &old_action);
     if(ret == -1){
-      perror("sigaction");
+      error(0, errno, "sigaction");
+      exitstatus = EX_OSERR;
       goto failure;
     }
     if(old_action.sa_handler != SIG_IGN){
       ret = sigaction(SIGTERM, &new_action, NULL);
       if(ret == -1){
-	perror("sigaction");
+	error(0, errno, "sigaction");
+	exitstatus = EX_OSERR;
 	goto failure;
       }
     }
@@ -235,7 +284,8 @@ int main(__attribute__((unused))int argc,
     goto failure;
   }
   if(splashy_command_pid == -1){
-    perror("fork");
+    error(0, errno, "fork");
+    exitstatus = EX_OSERR;
     goto failure;
   }
   /* Child */
@@ -243,7 +293,33 @@ int main(__attribute__((unused))int argc,
     if(not interrupted_by_signal){
       const char splashy_command[] = "/sbin/splashy_update";
       execl(splashy_command, splashy_command, prompt, (char *)NULL);
-      perror("execl");
+      int e = errno;
+      error(0, errno, "execl");
+      switch(e){
+      case EACCES:
+      case ENOENT:
+      case ENOEXEC:
+      case EINVAL:
+	_exit(EX_UNAVAILABLE);
+      case ENAMETOOLONG:
+      case E2BIG:
+      case ENOMEM:
+      case EFAULT:
+      case EIO:
+      case EMFILE:
+      case ENFILE:
+      case ETXTBSY:
+      default:
+	_exit(EX_OSERR);
+      case ENOTDIR:
+      case ELOOP:
+      case EISDIR:
+#ifdef ELIBBAD
+      case ELIBBAD:		/* Linux only */
+#endif
+      case EPERM:
+	_exit(EX_OSFILE);
+      }
     }
     free(prompt);
     _exit(EXIT_FAILURE);
@@ -268,7 +344,7 @@ int main(__attribute__((unused))int argc,
       goto failure;
     }
     if(ret == -1){
-      perror("waitpid");
+      error(0, errno, "waitpid");
       if(errno == ECHILD){
 	splashy_command_pid = 0;
       }
@@ -306,26 +382,42 @@ int main(__attribute__((unused))int argc,
 	 the real user ID (_mandos) */
       ret = setuid(geteuid());
       if(ret == -1){
-	perror("setuid");
+	error(0, errno, "setuid");
       }
       
       setsid();
       ret = chdir("/");
       if(ret == -1){
-	perror("chdir");
+	error(0, errno, "chdir");
       }
 /*       if(fork() != 0){ */
 /* 	_exit(EXIT_SUCCESS); */
 /*       } */
       ret = dup2(STDERR_FILENO, STDOUT_FILENO); /* replace stdout */
       if(ret == -1){
-	perror("dup2");
-	_exit(EXIT_FAILURE);
+	error(0, errno, "dup2");
+	_exit(EX_OSERR);
       }
-    
+      
       execl("/sbin/splashy", "/sbin/splashy", "boot", (char *)NULL);
-      perror("execl");
-      _exit(EXIT_FAILURE);
+      {
+	int e = errno;
+	error(0, errno, "execl");
+	switch(e){
+	case EACCES:
+	case ENOENT:
+	case ENOEXEC:
+	default:
+	  _exit(EX_UNAVAILABLE);
+	case ENAMETOOLONG:
+	case E2BIG:
+	case ENOMEM:
+	  _exit(EX_OSERR);
+	case ENOTDIR:
+	case ELOOP:
+	  _exit(EX_OSFILE);
+	}
+      }
     }
   }
   
@@ -336,17 +428,17 @@ int main(__attribute__((unused))int argc,
     ret = (int)TEMP_FAILURE_RETRY(sigaction(signal_received,
 					    &signal_action, NULL));
     if(ret == -1){
-      perror("sigaction");
+      error(0, errno, "sigaction");
     }
     do {
       ret = raise(signal_received);
     } while(ret != 0 and errno == EINTR);
     if(ret != 0){
-      perror("raise");
+      error(0, errno, "raise");
       abort();
     }
     TEMP_FAILURE_RETRY(pause());
   }
   
-  return EXIT_FAILURE;
+  return exitstatus;
 }
