@@ -2,8 +2,8 @@
 /*
  * Usplash - Read a password from usplash and output it
  * 
- * Copyright © 2008,2009 Teddy Hogeborn
- * Copyright © 2008,2009 Björn Påhlsson
+ * Copyright © 2008-2011 Teddy Hogeborn
+ * Copyright © 2008-2011 Björn Påhlsson
  * 
  * This program is free software: you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as
@@ -31,22 +31,25 @@
 #include <fcntl.h>		/* open(), O_WRONLY, O_RDONLY */
 #include <iso646.h>		/* and, or, not*/
 #include <errno.h>		/* errno, EINTR */
+#include <error.h>
 #include <sys/types.h>		/* size_t, ssize_t, pid_t, DIR, struct
 				   dirent */
 #include <stddef.h>		/* NULL */
 #include <string.h>		/* strlen(), memcmp() */
-#include <stdio.h>		/* asprintf(), perror() */
+#include <stdio.h>		/* asprintf()*/
 #include <unistd.h>		/* close(), write(), readlink(),
 				   read(), STDOUT_FILENO, sleep(),
 				   fork(), setuid(), geteuid(),
 				   setsid(), chdir(), dup2(),
 				   STDERR_FILENO, execv() */
 #include <stdlib.h>		/* free(), EXIT_FAILURE, realloc(),
-				   EXIT_SUCCESS, malloc(), _exit() */
-#include <stdlib.h>		/* getenv() */
+				   EXIT_SUCCESS, malloc(), _exit(),
+				   getenv() */
 #include <dirent.h>		/* opendir(), readdir(), closedir() */
 #include <inttypes.h>		/* intmax_t, strtoimax() */
 #include <sys/stat.h>		/* struct stat, lstat(), S_ISLNK */
+#include <sysexits.h>		/* EX_OSERR, EX_UNAVAILABLE */
+#include <argz.h>		/* argz_count(), argz_extract() */
 
 sig_atomic_t interrupted_by_signal = 0;
 int signal_received;
@@ -152,7 +155,7 @@ pid_t find_usplash(char **cmdline_r, size_t *cmdline_len_r){
   size_t cmdline_len = 0;
   DIR *proc_dir = opendir("/proc");
   if(proc_dir == NULL){
-    perror("opendir");
+    error(0, errno, "opendir");
     return -1;
   }
   errno = 0;
@@ -180,7 +183,7 @@ pid_t find_usplash(char **cmdline_r, size_t *cmdline_len_r){
       char *exe_link;
       ret = asprintf(&exe_link, "/proc/%s/exe", proc_ent->d_name);
       if(ret == -1){
-	perror("asprintf");
+	error(0, errno, "asprintf");
 	goto fail_find_usplash;
       }
       
@@ -192,7 +195,7 @@ pid_t find_usplash(char **cmdline_r, size_t *cmdline_len_r){
 	  free(exe_link);
 	  continue;
 	}
-	perror("lstat");
+	error(0, errno, "lstat");
 	free(exe_link);
 	goto fail_find_usplash;
       }
@@ -223,13 +226,13 @@ pid_t find_usplash(char **cmdline_r, size_t *cmdline_len_r){
 	ret = asprintf(&cmdline_filename, "/proc/%s/cmdline",
 		       proc_ent->d_name);
 	if(ret == -1){
-	  perror("asprintf");
+	  error(0, errno, "asprintf");
 	  goto fail_find_usplash;
 	}
 	cl_fd = open(cmdline_filename, O_RDONLY);
 	free(cmdline_filename);
 	if(cl_fd == -1){
-	  perror("open");
+	  error(0, errno, "open");
 	  goto fail_find_usplash;
 	}
       }
@@ -241,7 +244,7 @@ pid_t find_usplash(char **cmdline_r, size_t *cmdline_len_r){
 	if(cmdline_len + blocksize > cmdline_allocated){
 	  tmp = realloc(cmdline, cmdline_allocated + blocksize);
 	  if(tmp == NULL){
-	    perror("realloc");
+	    error(0, errno, "realloc");
 	    close(cl_fd);
 	    goto fail_find_usplash;
 	  }
@@ -252,7 +255,7 @@ pid_t find_usplash(char **cmdline_r, size_t *cmdline_len_r){
 	sret = read(cl_fd, cmdline + cmdline_len,
 		    cmdline_allocated - cmdline_len);
 	if(sret == -1){
-	  perror("read");
+	  error(0, errno, "read");
 	  close(cl_fd);
 	  goto fail_find_usplash;
 	}
@@ -260,14 +263,14 @@ pid_t find_usplash(char **cmdline_r, size_t *cmdline_len_r){
       } while(sret != 0);
       ret = close(cl_fd);
       if(ret == -1){
-	perror("close");
+	error(0, errno, "close");
 	goto fail_find_usplash;
       }
     }
     /* Close directory */
     ret = closedir(proc_dir);
     if(ret == -1){
-      perror("closedir");
+      error(0, errno, "closedir");
       goto fail_find_usplash;
     }
     /* Success */
@@ -297,9 +300,11 @@ int main(__attribute__((unused))int argc,
   size_t buf_len = 0;
   pid_t usplash_pid = -1;
   bool usplash_accessed = false;
+  int status = EXIT_FAILURE;	/* Default failure exit status */
   
   char *prompt = makeprompt();
   if(prompt == NULL){
+    status = EX_OSERR;
     goto failure;
   }
   
@@ -308,6 +313,7 @@ int main(__attribute__((unused))int argc,
   size_t cmdline_len = 0;
   usplash_pid = find_usplash(&cmdline, &cmdline_len);
   if(usplash_pid == 0){
+    status = EX_UNAVAILABLE;
     goto failure;
   }
   
@@ -319,23 +325,27 @@ int main(__attribute__((unused))int argc,
     sigemptyset(&new_action.sa_mask);
     ret = sigaddset(&new_action.sa_mask, SIGINT);
     if(ret == -1){
-      perror("sigaddset");
+      error(0, errno, "sigaddset");
+      status = EX_OSERR;
       goto failure;
     }
     ret = sigaddset(&new_action.sa_mask, SIGHUP);
     if(ret == -1){
-      perror("sigaddset");
+      error(0, errno, "sigaddset");
+      status = EX_OSERR;
       goto failure;
     }
     ret = sigaddset(&new_action.sa_mask, SIGTERM);
     if(ret == -1){
-      perror("sigaddset");
+      error(0, errno, "sigaddset");
+      status = EX_OSERR;
       goto failure;
     }
     ret = sigaction(SIGINT, NULL, &old_action);
     if(ret == -1){
       if(errno != EINTR){
-	perror("sigaction");
+	error(0, errno, "sigaction");
+	status = EX_OSERR;
       }
       goto failure;
     }
@@ -343,7 +353,8 @@ int main(__attribute__((unused))int argc,
       ret = sigaction(SIGINT, &new_action, NULL);
       if(ret == -1){
 	if(errno != EINTR){
-	  perror("sigaction");
+	  error(0, errno, "sigaction");
+	  status = EX_OSERR;
 	}
 	goto failure;
       }
@@ -351,7 +362,8 @@ int main(__attribute__((unused))int argc,
     ret = sigaction(SIGHUP, NULL, &old_action);
     if(ret == -1){
       if(errno != EINTR){
-	perror("sigaction");
+	error(0, errno, "sigaction");
+	status = EX_OSERR;
       }
       goto failure;
     }
@@ -359,7 +371,8 @@ int main(__attribute__((unused))int argc,
       ret = sigaction(SIGHUP, &new_action, NULL);
       if(ret == -1){
 	if(errno != EINTR){
-	  perror("sigaction");
+	  error(0, errno, "sigaction");
+	  status = EX_OSERR;
 	}
 	goto failure;
       }
@@ -367,7 +380,8 @@ int main(__attribute__((unused))int argc,
     ret = sigaction(SIGTERM, NULL, &old_action);
     if(ret == -1){
       if(errno != EINTR){
-	perror("sigaction");
+	error(0, errno, "sigaction");
+	status = EX_OSERR;
       }
       goto failure;
     }
@@ -375,7 +389,8 @@ int main(__attribute__((unused))int argc,
       ret = sigaction(SIGTERM, &new_action, NULL);
       if(ret == -1){
 	if(errno != EINTR){
-	  perror("sigaction");
+	  error(0, errno, "sigaction");
+	  status = EX_OSERR;
 	}
 	goto failure;
       }
@@ -386,7 +401,8 @@ int main(__attribute__((unused))int argc,
   /* Write command to FIFO */
   if(not usplash_write(&fifo_fd, "TIMEOUT", "0")){
     if(errno != EINTR){
-      perror("usplash_write");
+      error(0, errno, "usplash_write");
+      status = EX_OSERR;
     }
     goto failure;
   }
@@ -397,7 +413,8 @@ int main(__attribute__((unused))int argc,
   
   if(not usplash_write(&fifo_fd, "INPUTQUIET", prompt)){
     if(errno != EINTR){
-      perror("usplash_write");
+      error(0, errno, "usplash_write");
+      status = EX_OSERR;
     }
     goto failure;
   }
@@ -414,7 +431,8 @@ int main(__attribute__((unused))int argc,
   outfifo_fd = open("/dev/.initramfs/usplash_outfifo", O_RDONLY);
   if(outfifo_fd == -1){
     if(errno != EINTR){
-      perror("open");
+      error(0, errno, "open");
+      status = EX_OSERR;
     }
     goto failure;
   }
@@ -432,7 +450,8 @@ int main(__attribute__((unused))int argc,
       char *tmp = realloc(buf, buf_allocated + blocksize);
       if(tmp == NULL){
 	if(errno != EINTR){
-	  perror("realloc");
+	  error(0, errno, "realloc");
+	  status = EX_OSERR;
 	}
 	goto failure;
       }
@@ -443,7 +462,8 @@ int main(__attribute__((unused))int argc,
 		buf_allocated - buf_len);
     if(sret == -1){
       if(errno != EINTR){
-	perror("read");
+	error(0, errno, "read");
+	status = EX_OSERR;
       }
       TEMP_FAILURE_RETRY(close(outfifo_fd));
       goto failure;
@@ -457,7 +477,8 @@ int main(__attribute__((unused))int argc,
   ret = close(outfifo_fd);
   if(ret == -1){
     if(errno != EINTR){
-      perror("close");
+      error(0, errno, "close");
+      status = EX_OSERR;
     }
     goto failure;
   }
@@ -469,7 +490,8 @@ int main(__attribute__((unused))int argc,
   
   if(not usplash_write(&fifo_fd, "TIMEOUT", "15")){
     if(errno != EINTR){
-      perror("usplash_write");
+      error(0, errno, "usplash_write");
+      status = EX_OSERR;
     }
     goto failure;
   }
@@ -481,7 +503,8 @@ int main(__attribute__((unused))int argc,
   ret = close(fifo_fd);
   if(ret == -1){
     if(errno != EINTR){
-      perror("close");
+      error(0, errno, "close");
+      status = EX_OSERR;
     }
     goto failure;
   }
@@ -494,7 +517,8 @@ int main(__attribute__((unused))int argc,
       sret = write(STDOUT_FILENO, buf + written, buf_len - written);
       if(sret == -1){
 	if(errno != EINTR){
-	  perror("write");
+	  error(0, errno, "write");
+	  status = EX_OSERR;
 	}
 	goto failure;
       }
@@ -523,14 +547,14 @@ int main(__attribute__((unused))int argc,
   
   /* If usplash was never accessed, we can stop now */
   if(not usplash_accessed){
-    return EXIT_FAILURE;
+    return status;
   }
   
   /* Close FIFO */
   if(fifo_fd != -1){
     ret = (int)TEMP_FAILURE_RETRY(close(fifo_fd));
     if(ret == -1 and errno != EINTR){
-      perror("close");
+      error(0, errno, "close");
     }
     fifo_fd = -1;
   }
@@ -539,31 +563,19 @@ int main(__attribute__((unused))int argc,
   if(outfifo_fd != -1){
     ret = (int)TEMP_FAILURE_RETRY(close(outfifo_fd));
     if(ret == -1){
-      perror("close");
+      error(0, errno, "close");
     }
   }
   
-  /* Create argc and argv for new usplash*/
-  int cmdline_argc = 0;
-  char **cmdline_argv = malloc(sizeof(char *));
-  {
-    size_t position = 0;
-    while(position < cmdline_len){
-      char **tmp = realloc(cmdline_argv,
-			   (sizeof(char *)
-			    * (size_t)(cmdline_argc + 2)));
-      if(tmp == NULL){
-	perror("realloc");
-	free(cmdline_argv);
-	return EXIT_FAILURE;
-      }
-      cmdline_argv = tmp;
-      cmdline_argv[cmdline_argc] = cmdline + position;
-      cmdline_argc++;
-      position += strlen(cmdline + position) + 1;
-    }
-    cmdline_argv[cmdline_argc] = NULL;
+  /* Create argv for new usplash*/
+  char **cmdline_argv = malloc((argz_count(cmdline, cmdline_len) + 1)
+			       * sizeof(char *)); /* Count args */
+  if(cmdline_argv == NULL){
+    error(0, errno, "malloc");
+    return status;
   }
+  argz_extract(cmdline, cmdline_len, cmdline_argv); /* Create argv */
+  
   /* Kill old usplash */
   kill(usplash_pid, SIGTERM);
   sleep(2);
@@ -580,34 +592,38 @@ int main(__attribute__((unused))int argc,
        the real user ID (_mandos) */
     ret = setuid(geteuid());
     if(ret == -1){
-      perror("setuid");
+      error(0, errno, "setuid");
     }
     
     setsid();
     ret = chdir("/");
+    if(ret == -1){
+      error(0, errno, "chdir");
+      _exit(EX_OSERR);
+    }
 /*     if(fork() != 0){ */
 /*       _exit(EXIT_SUCCESS); */
 /*     } */
     ret = dup2(STDERR_FILENO, STDOUT_FILENO); /* replace our stdout */
     if(ret == -1){
-      perror("dup2");
-      _exit(EXIT_FAILURE);
+      error(0, errno, "dup2");
+      _exit(EX_OSERR);
     }
     
     execv(usplash_name, cmdline_argv);
     if(not interrupted_by_signal){
-      perror("execv");
+      error(0, errno, "execv");
     }
     free(cmdline);
     free(cmdline_argv);
-    _exit(EXIT_FAILURE);
+    _exit(EX_OSERR);
   }
   free(cmdline);
   free(cmdline_argv);
   sleep(2);
   if(not usplash_write(&fifo_fd, "PULSATE", NULL)){
     if(errno != EINTR){
-      perror("usplash_write");
+      error(0, errno, "usplash_write");
     }
   }
   
@@ -615,7 +631,7 @@ int main(__attribute__((unused))int argc,
   if(fifo_fd != -1){
     ret = (int)TEMP_FAILURE_RETRY(close(fifo_fd));
     if(ret == -1 and errno != EINTR){
-      perror("close");
+      error(0, errno, "close");
     }
     fifo_fd = -1;
   }
@@ -626,17 +642,17 @@ int main(__attribute__((unused))int argc,
     ret = (int)TEMP_FAILURE_RETRY(sigaction(signal_received,
 					    &signal_action, NULL));
     if(ret == -1){
-      perror("sigaction");
+      error(0, errno, "sigaction");
     }
     do {
       ret = raise(signal_received);
     } while(ret != 0 and errno == EINTR);
     if(ret != 0){
-      perror("raise");
+      error(0, errno, "raise");
       abort();
     }
     TEMP_FAILURE_RETRY(pause());
   }
   
-  return EXIT_FAILURE;
+  return status;
 }
